import { Rand } from 'pi_math/rand';
import { Tr } from "pi_ucore/db/mgr";
import { dropTwo, randomInt } from "./lottery";

import { DateMgr } from "pi_business_tool/date_mgr";
import { makeByJson, Prob } from 'pi_math/nest_prob';
import { Struct } from "pi_serialization/struct_mgr";
import { IStruct, read } from "pi_common_tools/util/db_tool";
import { roundCfg, getPrimaryKey, curRoundBigAward } from "../base/base";
import { getBaseCfg, getLuckDrawCfg, getChooseCfg } from "../base/cfg_tool";
import { Lottery as LotteryDB, BigChooseRecord } from "game_mj_lottery/server/lottery.struct";

/**
 * 获取随机组ID（根据配置的{"prob_p":[1,1,1],"prob_v":[9990101,9990102,9990103]}获取）
 * @param randomConfig 
 * @returns 
 */
const getRandomID = (randomConfig) => {
    // 声明一个接收值的变量
    let randomResult = [];
    let makeInfo: any = makeByJson(randomConfig);
    if (makeInfo) { new Prob(makeInfo.values, makeInfo.probs).select(new Rand(Math.floor(Math.random() * 0xffffffff)), randomResult) };
    return randomResult;
}

/**
 * 获取大奖的位置索引
 * @param roundId 轮次id
 * @param cost 代币消耗
 * @returns 大奖索引位置
 */
const bigIdx = (roundId: number, luck: number) => {
    const cfg = roundCfg(roundId);
    const min = Math.random() * 10000;
    const min_rate = luck >= cfg.cost_luck ? cfg.min_rate_change : cfg.min_rate;
    return min <= min_rate ? randomInt(cfg.min_min_value, cfg.min_max_value) : randomInt(cfg.max_min_value, cfg.max_max_value);
}

/**
 * 获取抽卡奖励和大奖信息
 * @param tr 
 * @param lotteryDB 抽卡记录db
 * @param items     待修改数据db
 * @returns 
 */
const _lottery = async (tr: Tr, lotteryDB: LotteryDB, items: [IStruct<Struct>, any, any][]) => {
    const getNums = lotteryDB.record.filter(v => v === 1).length;
    const isBig = (lotteryDB.bigIdx - getNums) === 1;
    if (isBig) {
        const curCfg = roundCfg(lotteryDB.curRound);
        let award = curCfg.BigAward;
        //当前轮次大奖不存在(记录自选轮次大奖次数)
        if (!curCfg.BigAward || curCfg.BigAward.length <= 0) {
            const chooseRecord = await read(tr, BigChooseRecord, lotteryDB.key);
            const chooseCfg = getChooseCfg(chooseRecord.curBigId);
            let num = chooseRecord.record.get(chooseRecord.curBigId) || 0;
            num += 1;
            chooseRecord.record.set(chooseRecord.curBigId, num);
            //超出次数重置
            if (chooseCfg.ChooseTimes && chooseCfg.ChooseTimes <= num) chooseRecord.curBigId = 0;
            items.push([BigChooseRecord, lotteryDB.key, chooseRecord]);
            award = chooseCfg.ChooseBigAward;
        }
        return {
            award,
            isBig: true
        }
    } else {
        const baseCfg = getBaseCfg();
        const lockCfg = getLuckDrawCfg(baseCfg.luck_draw_id);
        const award = await dropTwo(tr, getRandomID(lockCfg.randomAward));
        return {
            award,
            isBig: false
        }
    }
}

/**
 * 初始活动
 * @param tr 
 * @param rid   玩家rid
 * @param key   活动key
 * @param items 待修改数据
 * @param luck  幸运值
 * @returns 
 */
export const init = async (tr: Tr, rid: number, key: string, items: [IStruct<Struct>, any, any][], openTime: number, luck: number = 0, refHour = 0) => {
    const primaryKey = getPrimaryKey(key, rid);
    const db = await read(tr, LotteryDB, primaryKey, 3000, false, true);
    if (db) return false;
    const cfg = roundCfg(1);
    const baseCfg = getBaseCfg();
    const lockCfg = getLuckDrawCfg(baseCfg.luck_draw_id);
    const now = DateMgr.now();
    const dayTime = DateMgr.dailyZero(now) + (24 + refHour) * DateMgr.OneHourMs;
    const openCoolTime = DateMgr.dailyZero(openTime) + (lockCfg.cooling + refHour) * DateMgr.OneHourMs; //开始时间的刷新时间
    const coolTime = DateMgr.dailyZero(now) + (lockCfg.cooling + refHour) * DateMgr.OneHourMs;          //免费次数刷新时间
    const coolTimes = Math.floor((coolTime - openCoolTime) / DateMgr.OneDayMs) + 1
    const lotteryDB = new LotteryDB();
    lotteryDB.key = primaryKey;
    lotteryDB.allCount = 0;
    lotteryDB.dayCount = 0;
    lotteryDB.curRound = 1;
    lotteryDB.dayTime = dayTime;
    lotteryDB.coolTime = coolTime;
    lotteryDB.bigIdx = bigIdx(1, luck);
    lotteryDB.freeCount = lockCfg.freeCount * coolTimes;
    lotteryDB.record = new Array(Math.pow(cfg.area, 2)).fill(0);
    items.push([LotteryDB, primaryKey, lotteryDB]);
    return true;
}

/**
 * 
 * @param tr 
 * @param rid 
 * @param key 
 * @param items 
 * @param luck  幸运值
 * @returns 
 */
export const go_next_round = async (tr: Tr, rid: number, key: string, items: [IStruct<Struct>, any, any][], luck: number = 0) => {
    const primaryKey = getPrimaryKey(key, rid);
    const lotteryDB = await read(tr, LotteryDB, primaryKey, 3000, false, true);
    if (!lotteryDB) return false;
    //未抽完奖励无法进入下一轮
    const getNums = lotteryDB.record.filter(v => v === 1).length;
    if (lotteryDB.bigIdx > getNums) return false;
    const nextRound = lotteryDB.curRound + 1;
    const cfg = roundCfg(nextRound);
    lotteryDB.curRound = nextRound;
    lotteryDB.bigIdx = bigIdx(nextRound, luck);
    lotteryDB.record = new Array(Math.pow(cfg.area, 2)).fill(0);
    items.push([LotteryDB, primaryKey, lotteryDB]);
    return true;
}

/**
 * 抽卡
 * @param tr 
 * @param rid       玩家rid
 * @param key       活动key
 * @param items     待修改数据表
 * @param times     抽卡次数
 * @param lIdx      抽卡索引
 * @param refHour   日刷新时间
 * @returns 
 */
export const lottery = async (tr: Tr, rid: number, key: string, items: [IStruct<Struct>, any, any][], times = 1, lIdx?: number, refHour = 0) => {

    if (times < 1) return false;

    const primaryKey = getPrimaryKey(key, rid);
    const lotteryDB = await read(tr, LotteryDB, primaryKey, 3000, false, true);
    if (!lotteryDB) return false;

    const now = DateMgr.now();
    const baseCfg = getBaseCfg();
    const lockCfg = getLuckDrawCfg(baseCfg.luck_draw_id);

    //刷新日抽卡次数
    if (now >= lotteryDB.dayTime) {
        const dayTime = DateMgr.dailyZero(now) + (24 + refHour) * DateMgr.OneHourMs;  //下一天刷新时间
        lotteryDB.dayTime = dayTime;
        lotteryDB.dayCount = 0;
    }

    //刷新免费次数
    if (lockCfg.cooling && now >= lotteryDB.coolTime) {
        const coolTime = DateMgr.dailyZero(now) + (lockCfg.cooling + refHour) * DateMgr.OneHourMs;  //下一天刷新时间
        const coolTimes = Math.floor((coolTime - lotteryDB.coolTime) / DateMgr.OneDayMs);           //免费次数天数
        lotteryDB.coolTime = coolTime;
        lotteryDB.freeCount += (lockCfg.freeCount * coolTimes);
    }
    
    if (lockCfg.cooling) { //修复免费次数刷新时间
        const coolTime = DateMgr.dailyZero(now) + (lockCfg.cooling + refHour) * DateMgr.OneHourMs;
        if (lotteryDB.coolTime !== coolTime) lotteryDB.coolTime = coolTime;
    }

    if (lotteryDB.dayCount >= baseCfg.maxdrawCount) return false;

    // 已经抽卡次数
    const getNums = lotteryDB.record.filter(v => v === 1).length;

    //大奖已经获得
    if (getNums >= lotteryDB.bigIdx) return false;

    const cfg = roundCfg(lotteryDB.curRound);

    // 当前没有大奖
    const bAward = await curRoundBigAward(tr, rid, key, lotteryDB.curRound);

    if (!bAward) return false;

    // 卡池总数
    const allTime = Math.pow(cfg.area, 2);

    // 指定抽卡池
    if (times === 1 && (lIdx === 0 || !!lIdx)) {
        //位置不对
        if (lIdx < 0 || lIdx >= allTime) return false;
        // 当前已抽完
        if (lotteryDB.record[lIdx]) return false;
    }

    //剩余抽卡次数
    const maxTime = allTime - getNums;

    //抽中大奖剩余次数
    const bigTime = lotteryDB.bigIdx - getNums;

    //实际抽卡次数(大奖的话直接获得)
    let lTime = times >= bigTime ? maxTime : times;

    //实际消耗抽卡次数
    let lotteryTime = Math.min(times, bigTime);

    const awardMap: Map<number, { isBig: boolean, award: number[][] }> = new Map();

    //大奖是否已出(出了索引依次增加)
    let isBig = false;

    //指定抽卡
    if (times === 1 && (lIdx === 0 || lIdx)) {
        const data = await _lottery(tr, lotteryDB as LotteryDB, items);
        awardMap.set(lIdx, data);
        lotteryDB.record[lIdx] = 1;
        if (data.isBig) isBig = true;
        lTime -= 1;
    }

    while (lTime > 0) {
        const data = await _lottery(tr, lotteryDB as LotteryDB, items);
        if (data.isBig) isBig = true;
        let idx = -1;
        if (isBig) {
            idx = lotteryDB.record.findIndex(v => !v);
        } else {
            //随机索引
            const idxs = [];
            for (let i = 0; i < lotteryDB.record.length; i++) {
                if (lotteryDB.record[i]) continue;
                idxs.push(i);
            }
            const _idx = randomInt(0, idxs.length - 1);
            idx = idxs[_idx];
        }
        awardMap.set(idx, data);
        lotteryDB.record[idx] = 1;
        lTime -= 1;
    }

    const freeCount = lotteryTime > lotteryDB.freeCount ? lotteryDB.freeCount : lotteryTime;
    lotteryDB.freeCount -= freeCount;
    lotteryDB.dayCount += lotteryTime;
    lotteryDB.allCount += lotteryTime;
    items.push([LotteryDB, lotteryDB.key, lotteryDB]);

    return {
        awardMap,       //抽卡得到奖励信息
        lotteryTime,    //实际抽卡次数
        oneCost: lockCfg.cost,
        tenCost: lockCfg.tenCost,
        costTime: lotteryTime - freeCount  //付费抽卡次数
    }
}

/**
 * 选择当前轮次大奖
 * @param tr    
 * @param rid   玩家rid
 * @param key   活动key
 * @param bigId 自选id
 * @param items 待修改数据
 * @returns 
 */
export const selBigAward = async (tr: Tr, rid: number, key: string, bigId: number, items: [IStruct<Struct>, any, any][]) => {
    const chooseCfg = getChooseCfg(bigId);

    //错误
    if (!chooseCfg) return false;
    const primaryKey = getPrimaryKey(key, rid);
    const choose = await read(tr, BigChooseRecord, primaryKey);

    //一致
    if (choose.curBigId === bigId) return false;

    //超出上限
    const num = choose.record.get(bigId) || 0;
    if (chooseCfg.ChooseTimes && num >= chooseCfg.ChooseTimes) return false;

    choose.curBigId = bigId;

    items.push([BigChooseRecord, primaryKey, choose]);

    return true;
}

/**
 * 获取抽卡表信息
 * @param tr 
 * @param rid 玩家rid
 * @param key 活动key
 * @returns 
 */
export const getLotteryInfo = async (tr: Tr, rid: number, key: string) => {
    const primaryKey = getPrimaryKey(key, rid);
    const lotteryDB = await read(tr, LotteryDB, primaryKey);
    return lotteryDB;
}